@@ -1,10 +1,12 @@
 from datetime import datetime, timedelta
 import urllib.parse
+import json
+import os
 from .apicaller import APICaller
 
 # Currently only utilizing the Montana state Metrc API. Future plans include ability to choose endpoint in a more
 # effective manner.
-API_BASE_URL = 'https://api-mt.metrc.com'
+ENDPOINTS_FILE = os.path.join(os.path.dirname(__file__), 'endpoints.json')
 
 
 class Metrc:
@@ -20,6 +22,8 @@ class Metrc:
     user_api_key : str
         API key tied to an individual which allows for access control and permissions for various subsets of related
         tasks within the state traceability system.
+    state : str
+        Two-letter state abbreviation for loading correct endpoints.
 
     Attributes
     ----------
@@ -39,35 +43,43 @@ class Metrc:
     mipps : list[str]
         Facilities that are identified for manufacturing and processing.
     providers : list[str]
-        Facilities that are identified as cultivators.T
+        Facilities that are identified as cultivators.
     '''
 
-    def __init__(self, vendor_api_key, user_api_key):
-        # As states can dictate the features available to clients, future plans include the abiltiy to dynamically
-        # provide a list of valid endpoints.
-        self.endpoints = {'facilities': ['/facilities/v1'],
-                          'packages': ['/packages/v1/active', '/packages/v1/inactive'],
-                          'harvests': ['/harvests/v1/active', '/harvests/v1/inactive'],
-                          'lab_results': ['/labtests/v1/results'],
-                          'outgoing_transfers': ['/transfers/v1/outgoing'],
-                          'deliveries': ['/transfers/v1/{id}/deliveries'],
-                          'sales_receipts': ['/sales/v1/receipts/active'],
-                          'sales_transactions': ['/sales/v1/receipts/{id}']}
+    def __init__(self, vendor_api_key, user_api_key, state):
+        self.endpoints = self.load_endpoints(state)
+        self.api_base_url = self.endpoints['base_url'][0]
 
         self.vendor_api_key = vendor_api_key
         self.user_api_key = user_api_key
         self.api = APICaller(self.vendor_api_key, self.user_api_key)
         self.init_facilities()
 
+    def load_endpoints(self, state):
+        '''
+        Load the endpoints.json file and select endpoints from the state specified.
+
+        Parameters
+        ----------
+        state : str
+            Two-letter state abbreviation.
+
+        Returns
+        -------
+        dict
+            Dictionary of endpoint names as keys, URL endpoint stub as value.
+        '''
+        with open(ENDPOINTS_FILE, 'r') as f:
+            data = json.load(f)
+        return data[state]
+
     def init_facilities(self):
         '''
         Get and store facility data from Metrc. Additionally, create several facility-type specific lists for updating
         information relevant only to that specific type (i.e. sales transactions for dispensaries).
         '''
         print('Initializing facility list...')
-        ep = self.endpoints['facilities'][0]
-        url = f"{API_BASE_URL}{ep}"
-        self.facility_data = self.api.get_from_url_list([url])[0]
+        self.facility_data = self.get_basic('facilities_v1')
 
         self.facilities = [facility['License']['Number'] for facility in self.facility_data]
 
@@ -181,6 +193,15 @@ class Metrc:
 
         return period_tuples
 
+    def get_basic(self, endpoint, license_number=None):
+        ep = self.endpoints[endpoint][0]
+        if license_number:
+            url = f"{self.api_base_url}{ep}/?licenseNumber={license_number}"
+        else:
+            url = f"{self.api_base_url}{ep}"
+        data = self.api.get_from_url_list([url])[0]
+        return data
+
     def get_last_modified(self, endpoint, license_number, last_modified_start, last_modified_end, **kwargs):
         '''
         Retreive entities from an endpoint using last modified query parameters and flatten the list of lists of dicts.
@@ -200,6 +221,11 @@ class Metrc:
             End of time range for the overall query in ISO 8601 format.
         flatten : bool, optional
             Optional keyword argument for flattening any subdictionaries in the retrieved entities.
+
+        Returns
+        -------
+        data : list[dict]
+            Dictionary objects representing the target entity being requested.
         '''
         try:
             assert(license_number in self.facilities)
@@ -208,7 +234,7 @@ class Metrc:
 
         self._print_get_message(endpoint, license_number, last_modified_start, last_modified_end)
 
-        if endpoint == 'sales_transactions':
+        if endpoint == 'sales_v1_transactions':
             return self.get_sales_transactions(license_number=license_number,
                                                sales_date_start=last_modified_start,
                                                sales_date_end=last_modified_end,
@@ -223,6 +249,44 @@ class Metrc:
         else:
             return data
 
+    def get_ids(self, endpoint, license_number, ids):
+        '''
+        Retreive entities from an ID based endpoint. These return a single dictionary each, so the concurrent API calls
+        ultimately provide a list of dicts with no flattening necessary.
+
+        Parameters
+        ----------
+        endpoint : str
+            Endpoint/category of endpoints defined in the class attribute, endpoints (dict).
+        license_number : str
+            Facility license number to insert in query. User must have appropriate permissions for this facility granted
+            in Metrc to obain valid data.
+        ids : list[str]
+            IDs to insert in the endpoint or otherwise target in the query.
+
+        Returns
+        -------
+        data : list[dict]
+            Dictionary objects representing the target entity being requested.
+        '''
+        try:
+            assert(license_number in self.facilities)
+        except AssertionError:
+            print('Invalid license: User must have access to facility.')
+
+        print(f'Getting {len(ids)} {endpoint}...')
+        urls = []
+
+        for ep in self.endpoints[endpoint]:
+            for i in ids:
+                query_dict = {'licenseNumber': license_number}
+                query = urllib.parse.urlencode(query_dict)
+                ep_with_id = ep.replace('{id}', str(i))
+                urls.append(f"{self.api_base_url}{ep_with_id}?{query}")
+
+        data = self._get_from_urls(urls)
+        return data
+
     def get_sales_transactions(self, license_number, sales_date_start=None, sales_date_end=None, receipt_ids=None,
                                flatten=False):
         '''
@@ -270,33 +334,23 @@ class Metrc:
                 print('No receipt IDs provided: start and end date required.')
 
             print('No IDs provided when getting transactions...')
-            receipts = self.get_last_modified('sales_receipts', license_number, sales_date_start, sales_date_end)
+            receipts = self.get_last_modified('sales_v1_receipts_active', license_number, sales_date_start, sales_date_end)
             receipt_ids = [receipt['Id'] for receipt in receipts if isinstance(receipt, dict)]
             if not receipt_ids:
                 print("No receipts in time period specified. Skipping transaction lookup.")
                 return []
 
-        print(f'Getting {len(receipt_ids)} sales transactions...')
-        urls = []
-
-        for ep in self.endpoints['sales_transactions']:
-            for i in receipt_ids:
-                query_dict = {'licenseNumber': license_number}
-                query = urllib.parse.urlencode(query_dict)
-                ep_with_id = ep.replace('{id}', str(i))
-                urls.append(f"{API_BASE_URL}{ep_with_id}?{query}")
-
-        response = self._get_from_urls(urls)
+        transactions = self.get_ids('sales_v1_receipts_id', license_number, receipt_ids)
 
         # Flattening is useful for use in flat files, single database tables, and Elasticsearch indicies. Otherwise, the
         # returned dictionary includes a sub-list of transactions.
         if flatten:
             print('Flattening transactions...')
-            doc_list = []
+            flattened_transactions = []
 
             # Extract relevant information that would otherwise not be included when isolating an individual
             # transaction and append it.
-            for receipt in response:
+            for receipt in transactions:
                 receipt_info = {
                     'ReceiptNumber': receipt['ReceiptNumber'],
                     'SalesDateTime': receipt['SalesDateTime'],
@@ -311,11 +365,11 @@ class Metrc:
 
                 for transaction in receipt['Transactions']:
                     transaction.update(receipt_info)
-                    doc_list.append(transaction)
+                    flattened_transactions.append(transaction)
 
-            return doc_list
+            return flattened_transactions
         else:
-            return response
+            return transactions
 
     def _print_get_message(self, endpoint, license_number, start_date, end_date):
         print(f'Getting {endpoint} for {license_number} between {start_date} and {end_date}')
@@ -417,7 +471,7 @@ class Metrc:
         # Note: At the time of this writing, the GET sales receipt endpoint is broken returning a status 500 when
         # attempting to query by last modified start/end, or otherwise ignores the query parameters and defaults to the
         # last 24 hours.
-        if endpoint == 'sales_receipts':
+        if endpoint == 'sales_v1_receipts_active':
             query_start_term = 'salesDateStart'
             query_end_term = 'salesDateEnd'
         else:
@@ -430,7 +484,7 @@ class Metrc:
                               query_start_term: period[0],
                               query_end_term: period[1]}
                 query = urllib.parse.urlencode(query_dict)
-                urls.append(f"{API_BASE_URL}{ep}?{query}")
+                urls.append(f"{self.api_base_url}{ep}?{query}")
 
         return urls
 
@@ -440,4 +494,4 @@ if __name__ == '__main__':
 
     vendor_api_key = os.environ['METRC_VENDOR_API_KEY']
     user_api_key = os.environ['METRC_USER_API_KEY']
-    m = Metrc(vendor_api_key, user_api_key)
+    m = Metrc(vendor_api_key, user_api_key, 'mt')
